"""This task replaces C standard library includes with C++ equivalents. stdint.h
and assert.h are exceptions.
"""

import regex

from wpiformat.task import Task


class Header:
    def __init__(self, name, func_names=None, type_regexes=None, add_prefix=True):
        """Manages function and type names in standard library header.

        Keyword arguments:
        name -- header name string
        func_names -- set of function name strings (default set())
        type_regexes -- list of type regex strings (default [])
        add_prefix -- determines whether std:: prefix is added or removed
                      (default True)
        """
        if type_regexes is None:
            type_regexes = []
        if func_names is None:
            func_names = set()
        self.name = name
        self.func_names = func_names
        self.add_prefix = add_prefix

        regex_prefix = r""
        if add_prefix:
            self.prefix = r"std::"
            self.type_sub = r"std::\g<1>"
            regex_prefix = r""
        else:
            self.prefix = r""
            self.type_sub = r"\g<1>"
            regex_prefix = r"std::"

        if func_names:
            # Matches C standard library function uses. C standard library
            # function names are alphanumeric and start with a letter. If the
            # function name is preceded by a word character and a space, it's
            # a function definition instead of a usage.
            self.func_regex = regex.compile(
                # Preceded by nonword character and spaces, comma, arithmetic
                # operators, or "("
                r"(?:[^\w]\s+|,|\(|\+|-|\*|/)"
                + regex_prefix
                + r"([a-z][a-z0-9]*)"
                + r"(?:\()"  # C stdlib function name  # Followed by open parenthesis
            )
        else:
            self.func_regex = None

        if type_regexes:
            # Check for type uses

            # Preceded by beginning of file, "<" (template), " ", ",", "(", or
            # line separator and optional spaces
            lookbehind = r"(?<=^|\<| |,|\(|\n)"

            type_names = regex_prefix + r"(" + r"|".join(type_regexes) + r")"

            # Followed by optional spaces and ">", ")", ",", ";", pointer
            # asterisks, or ellipses
            lookahead = r"(?=(\s*(\>|\)|,|;|\*+|\.\.\.))|\s)"

            self.type_regex = regex.compile(lookbehind + type_names + lookahead)
        else:
            self.type_regex = None


class Stdlib(Task):
    def __init__(self):
        super().__init__()

        self.headers = []

        # assert is a macro, so it's ommitted to avoid prefixing with std::
        self.headers.append(Header("assert"))

        self.headers.append(
            Header(
                "ctype",
                {
                    "isalum",
                    "isalpha",
                    "isblank",
                    "iscntrl",
                    "isdigit",
                    "isgraph",
                    "islower",
                    "isprint",
                    "ispunct",
                    "isspace",
                    "isupper",
                    "isxdigit",
                    "tolower",
                    "toupper",
                },
            )
        )
        self.headers.append(Header("errno"))
        self.headers.append(Header("float"))
        self.headers.append(Header("limits"))
        self.headers.append(
            Header(
                "math",
                {
                    "cos",
                    "acos",
                    "cosh",
                    "acosh",
                    "sin",
                    "asin",
                    "asinh",
                    "tan",
                    "atan",
                    "atan2",
                    "atanh",
                    "exp",
                    "frexp",
                    "ldexp",
                    "log",
                    "log10",
                    "ilogb",
                    "log1p",
                    "log2",
                    "logb",
                    "modf",
                    "exp2",
                    "expm1",
                    "scalbl",
                    "scalbln",
                    "pow",
                    "sqrt",
                    "cbrt",
                    "hypot",
                    "erf",
                    "erfc",
                    "tgamma",
                    "lgamma",
                    "ceil",
                    "floor",
                    "fmod",
                    "trunc",
                    "round",
                    "lround",
                    "llround",
                    "rint",
                    "lrint",
                    "llrint",
                    "nearbyint",
                    "remainder",
                    "remquo",
                    "copysign",
                    "nan",
                    "nextafter",
                    "nexttoward",
                    "fdim",
                    "fmax",
                    "fmin",
                    "fma",
                    "fpclassify",
                    "abs",
                    "fabs",
                    "signbit",
                    "isfinite",
                    "isinf",
                    "isnan",
                    "isnormal",
                    "isgreater",
                    "isgreaterequal",
                    "isless",
                    "islessequal",
                    "islessgreater",
                    "isunordered",
                },
            )
        )
        self.headers.append(Header("setjmp", {"longjmp", "setjmp"}, ["jmp_buf"]))
        self.headers.append(
            Header("signal", {"signal", "raise"}, ["sig_atomic_t"], False)
        )
        self.headers.append(Header("stdarg", {"va_list"}))
        self.headers.append(
            Header("stddef", type_regexes=["(ptrdiff|max_align|nullptr)_t"])
        )

        # size_t isn't actually defined in stdint, but it fits best here for
        # removing the std:: prefix
        self.headers.append(
            Header(
                "stdint",
                type_regexes=["((u?int((_fast|_least)?(8|16|32|64)|max|ptr)|size)_t)"],
                add_prefix=False,
            )
        )

        self.headers.append(
            Header(
                "stdio",
                {
                    "remove",
                    "rename",
                    "rewind",
                    "tmpfile",
                    "tmpnam",
                    "fclose",
                    "fflush",
                    "fopen",
                    "freopen",
                    "fgetc",
                    "fgets",
                    "fputc",
                    "fputs",
                    "fread",
                    "fwrite",
                    "fgetpos",
                    "fseek",
                    "fsetpos",
                    "ftell",
                    "feof",
                    "ferror",
                    "setbuf",
                    "setvbuf",
                    "fprintf",
                    "snprintf",
                    "sprintf",
                    "vfprintf",
                    "vprintf",
                    "vsnprintf",
                    "vsprintf",
                    "printf",
                    "fscanf",
                    "sscanf",
                    "vfscanf",
                    "vscanf",
                    "vsscanf",
                    "scanf",
                    "getchar",
                    "gets",
                    "putc",
                    "putchar",
                    "puts",
                    "getc",
                    "ungetc",
                    "clearerr",
                    "perror",
                },
                ["FILE", "fpos_t"],
            )
        )
        self.headers.append(
            Header(
                "stdlib",
                {
                    "atof",
                    "atoi",
                    "atol",
                    "atoll",
                    "strtof",
                    "strtol",
                    "strtod",
                    "strtold",
                    "strtoll",
                    "strtoul",
                    "strtoull",
                    "rand",
                    "srand",
                    "free",
                    "calloc",
                    "malloc",
                    "realloc",
                    "abort",
                    "at_quick_exit",
                    "quick_exit",
                    "atexit",
                    "exit",
                    "getenv",
                    "system",
                    "_Exit",
                    "bsearch",
                    "qsort",
                    "llabs",
                    "labs",
                    "abs",
                    "lldiv",
                    "ldiv",
                    "div",
                    "mblen",
                    "btowc",
                    "wctomb",
                    "wcstombs",
                    "mbstowcs",
                },
                ["(l|ll)?div_t"],
            )
        )
        self.headers.append(
            Header(
                "string",
                {
                    "memcpy",
                    "memcmp",
                    "memchr",
                    "memmove",
                    "memset",
                    "strcpy",
                    "strncpy",
                    "strcat",
                    "strncat",
                    "strcmp",
                    "strncmp",
                    "strcoll",
                    "strchr",
                    "strrchr",
                    "strstr",
                    "strxfrm",
                    "strcspn",
                    "strrspn",
                    "strpbrk",
                    "strtok",
                    "strerror",
                    "strlen",
                },
            )
        )
        self.headers.append(
            Header(
                "time",
                {
                    "clock",
                    "asctime",
                    "ctime",
                    "difftime",
                    "gmtime",
                    "localtime",
                    "mktime",
                    "strftime",
                    "time",
                },
                ["(clock|time)_t"],
            )
        )

    @staticmethod
    def should_process_file(config_file, name):
        return config_file.is_cpp_file(name)

    def run_pipeline(self, config_file, name, lines):
        for header in self.headers:
            # Prepare include names
            before = ""
            after = ""
            if header.add_prefix:
                before = header.name + ".h"
                after = "c" + header.name
            else:
                before = "c" + header.name
                after = header.name + ".h"

            output_lines = []
            for line in lines.split("\n"):
                if "NOLINT" not in line:
                    output_lines.append(
                        line.replace(
                            "#include <" + before + ">", "#include <" + after + ">"
                        )
                    )
                else:
                    output_lines.append(line)
            lines = "\n".join(output_lines)

            if header.func_regex:
                lines = self.func_substitute(header, lines)

            if header.type_regex:
                lines = header.type_regex.sub(header.type_sub, lines)

        return lines, True

    @staticmethod
    def func_substitute(header, lines):
        """Returns modified lines and whether string changed.

        Keyword arguments:
        header -- Header object
        lines -- file contents string

        Returns modified file contents string
        """
        pos = 0
        while pos < len(lines):
            old_length = len(lines)

            # Check for function starting at "pos"
            match = header.func_regex.search(lines, pos)
            if not match:
                break

            # Set "pos" to after current match
            pos = match.start(1) + len(header.prefix) + len(match.group(1))

            # If function name is part of this header, substitute its name
            line = lines[match.start(1) : lines.find("\n", match.start(1))]
            if match.group(1) in header.func_names and "NOLINT" not in line:
                lines = (
                    lines[0 : match.start(1)]
                    + header.prefix
                    + match.group(1)
                    + lines[match.end(1) :]
                )
        return lines



"""This task removes extra newlines after the line containing "class"."""

import regex

from wpiformat.task import Task


class JavaClass(Task):
    @staticmethod
    def should_process_file(config_file, name):
        return name.endswith(".java")

    def run_pipeline(self, config_file, name, lines):
        linesep = super().get_linesep(lines)

        output = ""
        pos = 0

        # Match two or more line separators
        token_str = (
            r"/\*|\*/|//|"
            + linesep
            + r"|{"
            + linesep
            + r"(?P<extra>("
            + linesep
            + r")+)"
        )
        token_regex = regex.compile(token_str)

        in_multicomment = False
        in_comment = False

        for match in token_regex.finditer(lines):
            token = match.group()

            if token == "/*":
                in_multicomment = True
            elif token == "*/":
                in_multicomment = False
                in_comment = False
            elif token == "//":
                in_comment = True
            elif token == linesep:
                in_comment = False
            elif not in_multicomment and not in_comment:
                # Otherwise, the token is a class

                # Removes extra line separators
                output += lines[pos : match.span("extra")[0]]
                pos = match.span()[1]

        # Write rest of file if it wasn't all processed
        if pos < len(lines):
            output += lines[pos:]

        return output, True



"""This task disallows "using" declarations in global namespaces in headers."""

import regex

from wpiformat.task import Task


class UsingDeclaration(Task):
    @staticmethod
    def should_process_file(config_file, name):
        return config_file.is_cpp_header_file(name)

    def run_pipeline(self, config_file, name, lines):
        linesep = super().get_linesep(lines)
        format_succeeded = True

        # Tokenize file as brace opens, brace closes, and "using" declarations.
        # "using" declarations are scoped, so content inside any bracket pair is
        # considered outside the global namespace.
        token_regex = regex.compile(
            r"/\*|\*/|//|\\\\|\\\"|\"|\\'|'|" + linesep + r"|\{|\}|using\b"
        )

        brace_count = 0
        in_multicomment = False
        in_singlecomment = False
        in_string = False
        in_char = False
        for match in token_regex.finditer(lines):
            token = match.group()

            if token == "/*":
                if not in_singlecomment and not in_string and not in_char:
                    in_multicomment = True
            elif token == "*/":
                if not in_singlecomment and not in_string and not in_char:
                    in_multicomment = False
            elif token == "//":
                if not in_multicomment and not in_string and not in_char:
                    in_singlecomment = True
            elif in_singlecomment and linesep in token:
                # Ignore token if it's in a singleline comment. Only check it
                # for newlines to end the comment.
                in_singlecomment = False
            elif in_multicomment or in_singlecomment:
                # Tokens processed after this branch are ignored if they are in
                # comments
                continue
            elif token == '\\"':
                continue
            elif token == '"':
                if not in_char:
                    in_string = not in_string
            elif token == "\\'":
                continue
            elif token == "'":
                if not in_string:
                    in_char = not in_char
            elif in_string or in_char:
                # Tokens processed after this branch are ignored if they are in
                # double or single quotes
                continue
            elif token == "{":
                brace_count += 1
            elif token == "}":
                brace_count -= 1
            elif token.startswith("using"):
                if brace_count == 0:
                    linenum = lines.count(linesep, 0, match.start()) + 1
                    if "NOLINT" not in lines.splitlines()[linenum - 1]:
                        format_succeeded = False

                        # Extract using declaration
                        using_decl = lines[
                            match.start() : lines.find(";", match.start()) + 1
                        ]

                        print(f"{name}: {linenum}: '{using_decl}' in global namespace")

        return lines, format_succeeded



"""This task makes include guards follow the style stipulated by the Google
style guide.
"""

import os
import regex

from enum import Enum
from wpiformat.task import Task


class State(Enum):
    FINDING_IFNDEF = 1
    FINDING_ENDIF = 2
    DONE = 3


class IncludeGuard(Task):
    @staticmethod
    def should_process_file(config_file, name):
        return config_file.is_header_file(name)

    def run_pipeline(self, config_file, name, lines):
        linesep = super().get_linesep(lines)
        lines_list = lines.split(linesep)
        output_list = lines_list

        state = State.FINDING_IFNDEF
        ifndef_regex = regex.compile(r"#ifndef \w+", regex.ASCII)
        define_regex = regex.compile(r"#define \w+", regex.ASCII)

        if_preproc_count = 0
        for i in range(len(lines_list)):
            if state == State.FINDING_IFNDEF:
                if lines_list[i].lstrip().startswith("#ifndef ") and lines_list[
                    i + 1
                ].lstrip().startswith("#define "):
                    state = State.FINDING_ENDIF

                    guard = self.make_include_guard(config_file, name)
                    output_list[i] = ifndef_regex.sub("#ifndef " + guard, lines_list[i])
                    output_list[i + 1] = define_regex.sub(
                        "#define " + guard, lines_list[i + 1]
                    )
                    if_preproc_count += 1
                elif lines_list[i].lstrip().startswith("#pragma once"):
                    state = State.DONE
            elif state == State.FINDING_ENDIF:
                if "#if" in lines_list[i]:
                    if_preproc_count += 1
                elif "#endif" in lines_list[i]:
                    if_preproc_count -= 1

                if if_preproc_count == 0:
                    state = State.DONE
                    output_list[i] = "#endif  // " + guard
                else:
                    output_list[i] = lines_list[i]
            else:
                output_list[i] = lines_list[i]

        # If include guard not found
        if state == State.FINDING_IFNDEF:
            print(
                "Error: " + name + ": doesn't contain include guard or '#pragma once'"
            )
            return lines, False

        output = linesep.join(output_list).rstrip() + linesep
        return output, True

    def make_include_guard(self, config_file, name):
        """Returns properly formatted include guard based on repository root and
        file name.

        Keyword arguments:
        config_file -- Config object
        name -- file name string
        """
        repo_root_name_override = config_file.group("repoRootNameOverride")

        repo_root = super().get_repo_root()
        guard_root = os.path.relpath(name, repo_root)
        if not repo_root_name_override:
            guard_path = os.path.basename(repo_root) + os.sep
        else:
            guard_path = repo_root_name_override[0] + os.sep
        include_roots = config_file.group("includeGuardRoots")

        if include_roots:
            prefix = ""
            for include_root in include_roots:
                if guard_root.startswith(include_root) and len(include_root) > len(
                    prefix
                ):
                    prefix = include_root
            guard_path += guard_root[len(prefix) :]
            return (regex.sub(r"[^a-zA-Z0-9]", "_", guard_path).upper() + "_").lstrip(
                "_"
            )

        # No include guard roots matched, so append full name
        guard_path += guard_root
        return regex.sub(r"[^a-zA-Z0-9]", "_", guard_path).upper() + "_"



import os

from .test_tasktest import *
from wpiformat.bracecomment import BraceComment


def test_bracecomment():
    test = TaskTest(BraceComment())

    # Empty anonymous namespace
    test.add_input("./Test.h", "namespace {" + os.linesep + "}// comment" + os.linesep)
    test.add_output("namespace {" + os.linesep + "}  // namespace" + os.linesep, True)

    # Anonymous namespace containing comment
    test.add_input(
        "./Test.h",
        "namespace {"
        + os.linesep
        + "  // comment"
        + os.linesep
        + "}// comment"
        + os.linesep,
    )
    test.add_output(
        "namespace {"
        + os.linesep
        + "  // comment"
        + os.linesep
        + "}  // namespace"
        + os.linesep,
        True,
    )

    # namespace
    test.add_input(
        "./Test.h",
        "namespace hal {"
        + os.linesep
        + "  // comment"
        + os.linesep
        + "}// comment"
        + os.linesep,
    )
    test.add_output(
        "namespace hal {"
        + os.linesep
        + "  // comment"
        + os.linesep
        + "}  // namespace hal"
        + os.linesep,
        True,
    )

    # namespace with leftover input
    test.add_input(
        "./Test.h",
        "// comment before namespace"
        + os.linesep
        + "namespace hal {"
        + os.linesep
        + "  // comment"
        + os.linesep
        + "}// comment"
        + os.linesep
        + "// comment after namespace"
        + os.linesep,
    )
    test.add_output(
        "// comment before namespace"
        + os.linesep
        + "namespace hal {"
        + os.linesep
        + "  // comment"
        + os.linesep
        + "}  // namespace hal"
        + os.linesep
        + "// comment after namespace"
        + os.linesep,
        True,
    )

    # Braces within namespace
    test.add_input(
        "./Test.h",
        "namespace {"
        + os.linesep
        + os.linesep
        + "struct AnalogGyro {"
        + os.linesep
        + "  HAL_AnalogInputHandle handle;"
        + os.linesep
        + "  double voltsPerDegreePerSecond;"
        + os.linesep
        + "  double offset;"
        + os.linesep
        + "  int32_t center;"
        + os.linesep
        + "}"
        + os.linesep
        + os.linesep
        + "}"
        + os.linesep,
    )
    test.add_output(
        "namespace {"
        + os.linesep
        + os.linesep
        + "struct AnalogGyro {"
        + os.linesep
        + "  HAL_AnalogInputHandle handle;"
        + os.linesep
        + "  double voltsPerDegreePerSecond;"
        + os.linesep
        + "  double offset;"
        + os.linesep
        + "  int32_t center;"
        + os.linesep
        + "}"
        + os.linesep
        + os.linesep
        + "}  // namespace"
        + os.linesep,
        True,
    )

    # extern "C"
    test.add_input(
        "./Test.h",
        'extern "C" {'
        + os.linesep
        + "    // nothing"
        + os.linesep
        + "}// comment"
        + os.linesep,
    )
    test.add_output(
        'extern "C" {'
        + os.linesep
        + "    // nothing"
        + os.linesep
        + '}  // extern "C"'
        + os.linesep,
        True,
    )

    # Nested brackets should be handled properly
    test.add_input(
        "./Test.cpp",
        'extern "C" {'
        + os.linesep
        + "void func() {"
        + os.linesep
        + "  if (1) {"
        + os.linesep
        + "  } else if (1) {"
        + os.linesep
        + "  } else {"
        + os.linesep
        + "  }"
        + os.linesep
        + "}"
        + os.linesep
        + '}  // extern "C"'
        + os.linesep,
    )
    test.add_latest_input_as_output(True)

    # Nested brackets on same line
    test.add_input(
        "./Test.cpp",
        "namespace wpi {"
        + os.linesep
        + "{{}}"
        + os.linesep
        + "}  // namespace java"
        + os.linesep,
    )
    test.add_output(
        "namespace wpi {"
        + os.linesep
        + "{{}}"
        + os.linesep
        + "}  // namespace wpi"
        + os.linesep,
        True,
    )

    # Handle single-line statements correctly
    test.add_input("./Test.cpp", "namespace hal { Type typeName; }" + os.linesep)
    test.add_output(
        "namespace hal { Type typeName; }  // namespace hal" + os.linesep, True
    )

    # Two incorrect comments
    test.add_input(
        "./Test.h",
        "namespace {"
        + os.linesep
        + "    // nothing"
        + os.linesep
        + "}// comment"
        + os.linesep
        + "namespace Name {"
        + os.linesep
        + "    // nothing"
        + os.linesep
        + "}"
        + os.linesep,
    )
    test.add_output(
        "namespace {"
        + os.linesep
        + "    // nothing"
        + os.linesep
        + "}  // namespace"
        + os.linesep
        + "namespace Name {"
        + os.linesep
        + "    // nothing"
        + os.linesep
        + "}  // namespace Name"
        + os.linesep,
        True,
    )

    # Don't touch correct comment
    test.add_input(
        "./Test.h",
        "namespace {"
        + os.linesep
        + "    // nothing"
        + os.linesep
        + "}  // namespace"
        + os.linesep,
    )
    test.add_latest_input_as_output(True)

    # Handle braces in comments properly
    test.add_input(
        "./Path.h",
        "#ifndef ALLWPILIB_WPI_PATH_H_"
        + os.linesep
        + "#define ALLWPILIB_WPI_PATH_H_"
        + os.linesep
        + os.linesep
        + "namespace wpi {"
        + os.linesep
        + "namespace sys {"
        + os.linesep
        + "namespace path {"
        + os.linesep
        + os.linesep
        + "/// @{"
        + os.linesep
        + os.linesep
        + "}  // end namespace path"
        + os.linesep
        + "}  // namespace sys"
        + os.linesep
        + "}  // namespace wpi"
        + os.linesep
        + os.linesep
        + "#endif  // ALLWPILIB_WPI_PATH_H_"
        + os.linesep,
    )
    test.add_output(
        "#ifndef ALLWPILIB_WPI_PATH_H_"
        + os.linesep
        + "#define ALLWPILIB_WPI_PATH_H_"
        + os.linesep
        + os.linesep
        + "namespace wpi {"
        + os.linesep
        + "namespace sys {"
        + os.linesep
        + "namespace path {"
        + os.linesep
        + os.linesep
        + "/// @{"
        + os.linesep
        + os.linesep
        + "}  // namespace path"
        + os.linesep
        + "}  // namespace sys"
        + os.linesep
        + "}  // namespace wpi"
        + os.linesep
        + os.linesep
        + "#endif  // ALLWPILIB_WPI_PATH_H_"
        + os.linesep,
        True,
    )

    test.run(OutputType.FILE)



import os

from .test_tasktest import *
from wpiformat.whitespace import Whitespace


def test_whitespace():
    test = TaskTest(Whitespace())

    file_appendix = (
        "#pragma once"
        + os.linesep
        + os.linesep
        + "#include <iostream>"
        + os.linesep
        + os.linesep
        + "int main() {"
        + os.linesep
        + '  std::cout << "Hello World!";'
        + os.linesep
        + "}"
        + os.linesep
    )

    # Empty file
    test.add_input("./Test.h", "")
    test.add_output("", True)

    # No trailing whitespace
    test.add_input("./Test.h", file_appendix)
    test.add_output(file_appendix, True)

    # Two spaces trailing
    test.add_input(
        "./Test.h",
        "#pragma once"
        + os.linesep
        + os.linesep
        + "#include <iostream>"
        + os.linesep
        + os.linesep
        + "int main() {  "
        + os.linesep
        + '  std::cout << "Hello World!";  '
        + os.linesep
        + "}"
        + os.linesep,
    )
    test.add_output(file_appendix, True)

    # Two tabs trailing
    test.add_input(
        "./Test.h",
        "#pragma once"
        + os.linesep
        + os.linesep
        + "#include <iostream>"
        + os.linesep
        + os.linesep
        + "int main() {\t\t"
        + os.linesep
        + '  std::cout << "Hello World!";\t\t'
        + os.linesep
        + "}"
        + os.linesep,
    )
    test.add_output(file_appendix, True)

    test.run(OutputType.FILE)



import os

from .test_tasktest import *
from wpiformat.cidentlist import CIdentList


def test_cidentlist():
    test = TaskTest(CIdentList())

    # Main.cpp: signature for C++ function
    test.add_input(
        "./Main.cpp",
        "int main() {" + os.linesep + "  return 0;" + os.linesep + "}" + os.linesep,
    )
    test.add_latest_input_as_output(True)

    # Main.cpp: signature for C function in extern "C" block
    test.add_input(
        "./Main.cpp",
        'extern "C" {'
        + os.linesep
        + "int main() {"
        + os.linesep
        + "  return 0;"
        + os.linesep
        + "}"
        + os.linesep
        + "}"
        + os.linesep,
    )
    test.add_output(
        'extern "C" {'
        + os.linesep
        + "int main(void) {"
        + os.linesep
        + "  return 0;"
        + os.linesep
        + "}"
        + os.linesep
        + "}"
        + os.linesep,
        True,
    )

    # Main.cpp: signature for C function marked extern "C"
    test.add_input(
        "./Main.cpp",
        'extern "C" int main() {'
        + os.linesep
        + "  return 0;"
        + os.linesep
        + "}"
        + os.linesep,
    )
    test.add_output(
        'extern "C" int main(void) {'
        + os.linesep
        + "  return 0;"
        + os.linesep
        + "}"
        + os.linesep,
        True,
    )

    # Main.cpp: extern "C++" function nested in extern "C" block
    test.add_input(
        "./Main.cpp",
        'extern "C" {'
        + os.linesep
        + 'extern "C++" int main() {'
        + os.linesep
        + "  return 0;"
        + os.linesep
        + "}"
        + os.linesep
        + "}"
        + os.linesep,
    )
    test.add_latest_input_as_output(True)

    # Main.c: signature for C function
    test.add_input(
        "./Main.c",
        "int main() {" + os.linesep + "  return 0;" + os.linesep + "}" + os.linesep,
    )
    test.add_output(
        "int main(void) {" + os.linesep + "  return 0;" + os.linesep + "}" + os.linesep,
        True,
    )

    # Main.c: signature for C++ function in extern "C++" block
    test.add_input(
        "./Main.c",
        'extern "C++" {'
        + os.linesep
        + "int main() {"
        + os.linesep
        + "  return 0;"
        + os.linesep
        + "}"
        + os.linesep
        + "}"
        + os.linesep,
    )
    test.add_latest_input_as_output(True)

    # Main.c: signature for C++ function marked extern "C++"
    test.add_input(
        "./Main.c",
        'extern "C++" int main() {'
        + os.linesep
        + "  return 0;"
        + os.linesep
        + "}"
        + os.linesep,
    )
    test.add_latest_input_as_output(True)

    # Main.c: extern "C" function nested in extern "C++" block
    test.add_input(
        "./Main.c",
        'extern "C++" {'
        + os.linesep
        + 'extern "C" int main() {'
        + os.linesep
        + "  return 0;"
        + os.linesep
        + "}"
        + os.linesep
        + "}"
        + os.linesep,
    )
    test.add_output(
        'extern "C++" {'
        + os.linesep
        + 'extern "C" int main(void) {'
        + os.linesep
        + "  return 0;"
        + os.linesep
        + "}"
        + os.linesep
        + "}"
        + os.linesep,
        True,
    )

    # Don't match function calls
    test.add_input(
        "./Main.c",
        "int main() {"
        + os.linesep
        + "  foo();"
        + os.linesep
        + "  return 0;"
        + os.linesep
        + "}"
        + os.linesep,
    )
    test.add_output(
        "int main(void) {"
        + os.linesep
        + "  foo();"
        + os.linesep
        + "  return 0;"
        + os.linesep
        + "}"
        + os.linesep,
        True,
    )

    # Don't match function calls with return (return is a keyword not a return
    # type)
    test.add_input(
        "./Main.c",
        "int main() {" + os.linesep + "  return foo();" + os.linesep + "}" + os.linesep,
    )
    test.add_output(
        "int main(void) {"
        + os.linesep
        + "  return foo();"
        + os.linesep
        + "}"
        + os.linesep,
        True,
    )

    # Match function prototypes
    test.add_input(
        "./Main.c",
        "int main();"
        + os.linesep
        + os.linesep
        + "int main() {"
        + os.linesep
        + "  foo();"
        + os.linesep
        + "  return 0;"
        + os.linesep
        + "}"
        + os.linesep,
    )
    test.add_output(
        "int main(void);"
        + os.linesep
        + os.linesep
        + "int main(void) {"
        + os.linesep
        + "  foo();"
        + os.linesep
        + "  return 0;"
        + os.linesep
        + "}"
        + os.linesep,
        True,
    )

    # Make sure leaving extern block resets extern language type of
    # parent block
    test.add_input(
        "./Main.c",
        'extern "C++" {'
        + os.linesep
        + 'extern "C" int main() {'
        + os.linesep
        + "  return 0;"
        + os.linesep
        + "}"
        + os.linesep
        + "int func() {"
        + os.linesep
        + "  return 0;"
        + os.linesep
        + "}"
        + os.linesep
        + "}"
        + os.linesep,
    )
    test.add_output(
        'extern "C++" {'
        + os.linesep
        + 'extern "C" int main(void) {'
        + os.linesep
        + "  return 0;"
        + os.linesep
        + "}"
        + os.linesep
        + "int func() {"
        + os.linesep
        + "  return 0;"
        + os.linesep
        + "}"
        + os.linesep
        + "}"
        + os.linesep,
        True,
    )

    # Don't match lambda function that takes no arguments
    test.add_input(
        "./Main.cpp",
        'extern "C" {'
        + os.linesep
        + os.linesep
        + "HAL_Bool HAL_Initialize(int32_t timeout, int32_t mode) {"
        + os.linesep
        + "  std::atexit([]() {"
        + os.linesep
        + "    // Unregister our new data condition variable."
        + os.linesep
        + "  });"
        + os.linesep
        + "}"
        + os.linesep
        + os.linesep
        + '}  // extern "C"'
        + os.linesep,
    )
    test.add_latest_input_as_output(True)

    # Don't match a function call within a #ifdef
    test.add_input(
        "./Main.c",
        "ES_Event Elevator_Service_Run(ES_Event event) {"
        + os.linesep
        + "#ifdef USE_TATTLETALE"
        + os.linesep
        + "    ES_Tail(); // trace call stack end"
        + os.linesep
        + "#endif"
        + os.linesep
        + "}"
        + os.linesep,
    )
    test.add_latest_input_as_output(True)

    test.add_input(
        "./Timer.hpp",
        'extern "C" void Timer1IntHandler();'
        + os.linesep
        + os.linesep
        + "class Timer {"
        + os.linesep
        + "public:"
        + os.linesep
        + "    void Set(uint32_t newTime);"
        + os.linesep
        + os.linesep
        + "    void Start();"
        + os.linesep
        + os.linesep
        + "    void Stop();"
        + os.linesep
        + os.linesep
        + "    uint16_t GetID() const;"
        + os.linesep
        + os.linesep
        + "    static uint32_t GetTime();"
        + os.linesep
        + os.linesep
        + "private:"
        + os.linesep
        + "    friend void Timer1IntHandler();"
        + os.linesep
        + "};"
        + os.linesep,
    )
    test.add_output(
        'extern "C" void Timer1IntHandler(void);'
        + os.linesep
        + os.linesep
        + "class Timer {"
        + os.linesep
        + "public:"
        + os.linesep
        + "    void Set(uint32_t newTime);"
        + os.linesep
        + os.linesep
        + "    void Start();"
        + os.linesep
        + os.linesep
        + "    void Stop();"
        + os.linesep
        + os.linesep
        + "    uint16_t GetID() const;"
        + os.linesep
        + os.linesep
        + "    static uint32_t GetTime();"
        + os.linesep
        + os.linesep
        + "private:"
        + os.linesep
        + "    friend void Timer1IntHandler();"
        + os.linesep
        + "};"
        + os.linesep,
        True,
    )

    # Ensure comments with } in them don't mess up brace stack
    test.add_input(
        "./Test.cpp",
        "void func() {" + os.linesep + "  // closing }" + os.linesep + "}" + os.linesep,
    )
    test.add_latest_input_as_output(True)

    # Ensure nested comments don't mess up brace stack
    test.add_input("./Test.cpp", "// { /* */ }" + os.linesep)
    test.add_latest_input_as_output(True)
    test.add_input("./Test.cpp", "{ // /* */ }" + os.linesep)
    test.add_latest_input_as_output(False)
    test.add_input("./Test.cpp", "{ /* // */ }" + os.linesep)
    test.add_latest_input_as_output(True)
    test.add_input("./Test.cpp", "{ /* */ // }" + os.linesep)
    test.add_latest_input_as_output(False)
    test.add_input("./Test.cpp", "{ // /* // */ }" + os.linesep)
    test.add_latest_input_as_output(False)

    # Ensure popping too many braces doesn't crash
    test.add_input("./Test.cpp", "}" + os.linesep)
    test.add_latest_input_as_output(False)

    # Ensure comments inside quoted string don't mess up brace stack
    test.add_input(
        "./Test.cpp",
        "void func() {"
        + os.linesep
        + '  // "//"'
        + os.linesep
        + '  if (!query.startswith("//")) {'
        + os.linesep
        + "    return;"
        + os.linesep
        + "  }"
        + os.linesep
        + "}"
        + os.linesep,
    )
    test.add_latest_input_as_output(True)

    # Ensure braces in double quotes don't mess up brace stack
    test.add_input("./Test.cpp", "void func() { std::cout << '{'; }")
    test.add_latest_input_as_output(True)

    # Ensure braces in single quotes don't mess up brace stack
    test.add_input("./Test.cpp", 'void func() { std::cout << "{"; }')
    test.add_latest_input_as_output(True)

    # Ensure single quote within double quotes doesn't mess up brace stack
    test.add_input("./Test.cpp", 'void func() { std::cout << "\'"; }')
    test.add_latest_input_as_output(True)

    # Ensure double quote within single quotes doesn't mess up brace stack
    test.add_input("./Test.cpp", "void func() { std::cout << '\"'; }")
    test.add_latest_input_as_output(True)

    # Ensure escaped double quote doesn't mess up brace stack
    test.add_input("./Test.cpp", 'void func() { std::cout << "\\""; }')
    test.add_latest_input_as_output(True)

    # Ensure escaped single quote doesn't mess up brace stack
    test.add_input("./Test.cpp", "void func() { std::cout << '\\''; }")
    test.add_latest_input_as_output(True)

    # Ensure escaped backslash isn't considered as an escaped single quote
    test.add_input("./Test.cpp", "void func() { std::cout << '\\\\'; }")
    test.add_latest_input_as_output(True)

    # Ensure extern "C" match containing a linesep within a singleline comment
    # still ends the comment
    test.add_input(
        "./Test.cpp",
        'extern "C" {}  // extern "C"'
        + os.linesep
        + "namespace {"
        + os.linesep
        + "}  // namespace"
        + os.linesep,
    )
    test.add_latest_input_as_output(True)

    # Ensure extern "C" with brace on next line gets matched
    test.add_input(
        "./Test.cpp",
        'extern "C"'
        + os.linesep
        + "{"
        + os.linesep
        + "  void func() {}"
        + os.linesep
        + '}  // extern "C"'
        + os.linesep,
    )
    test.add_output(
        'extern "C"'
        + os.linesep
        + "{"
        + os.linesep
        + "  void func(void) {}"
        + os.linesep
        + '}  // extern "C"'
        + os.linesep,
        True,
    )

    # Test logic for deduplicating braces within #ifdef
    test.add_input(
        "./Test.cpp",
        "void func() {"
        + os.linesep
        + "#ifdef _WIN32"
        + os.linesep
        + "  if (errno == WSAEWOULDBLOCK) {"
        + os.linesep
        + "#else"
        + os.linesep
        + "  if (errno == EWOULDBLOCK) {"
        + os.linesep
        + "#endif"
        + os.linesep
        + "  }"
        + os.linesep
        + "}"
        + os.linesep,
    )
    test.add_latest_input_as_output(True)

    # Ensure extern "C" function with pointer return type gets matched
    test.add_input(
        "./Test.cpp",
        'extern "C" void* func() {}' + os.linesep,
    )
    test.add_output(
        'extern "C" void* func(void) {}' + os.linesep,
        True,
    )

    test.run(OutputType.FILE)



"""This task sorts C/C++ includes."""

import os
import regex

from wpiformat.task import Task


class IncludeOrder(Task):
    def __init__(self):
        super().__init__()

        # There are 5 header groups:
        # 0. Related headers
        # 1. C system headers (includes standard library headers)
        # 2. C++ system headers (includes standard library headers)
        # 3. Other library headers
        # 4. Project headers
        #
        # See comments below for how headers are classified.

        # Header type 0: Related headers
        # Base name of include matches base name of current file

        # Header type 1: C standard library headers
        # From https://en.cppreference.com/w/cpp/header
        self.c_std = [
            # C compatibility headers
            "assert.h",
            "ctype.h",
            "errno.h",
            "fenv.h",
            "float.h",
            "inttypes.h",
            "limits.h",
            "locale.h",
            "math.h",
            "setjmp.h",
            "signal.h",
            "stdarg.h",
            "stddef.h",
            "stdint.h",
            "stdio.h",
            "stdlib.h",
            "string.h",
            "time.h",
            "uchar.h",
            "wchar.h",
            "wctype.h",
            # Special C compatibility headers
            "stdatomic.h",
            # Empty C headers
            "ccomplex",
            "complex.h",
            "ctgmath",
            "tgmath.h",
            # Meaningless C headers
            "ciso646",
            "cstdalign",
            "cstdbool",
            "iso646.h",
            "stdalign.h",
            "stdbool.h",
            # Unsupported C headers
            "stdatomic.h",
            "stdnoreturn.h",
            "threads.h",
        ]

        # Header type 1: C system headers
        self.c_sys_regex = regex.compile(r"<[a-z][A-Za-z0-9/_-]*\.h>")

        # Header type 2: C++ standard library headers
        # With the exception of a few C standard library headers above, they are
        # the only headers which use only lowercase and underscores, and don't
        # have a ".h" suffix
        self.cpp_std_regex = regex.compile(r"(<|\")[a-z0-9_]+(>|\")")

        # Header type 3: Other library headers
        # They use angle brackets (open_bracket group is angle bracket)
        #
        # Header type 4: Project headers
        # They use double quotes (all other headers)
        self.header_regex = regex.compile(
            r"(?P<comment>//\s*)?"
            r"\#include\s*"
            r"(?P<header>"
            r"(?P<open_bracket><|\")"
            r"(?P<name>[^>\"]*)"
            r"(?P<close_bracket>>|\"))"
            r"(?P<postfix>.*)$"
        )

    @staticmethod
    def should_process_file(config_file, name):
        return config_file.is_c_file(name) or config_file.is_cpp_file(name)

    def classify_header(self, config_file, include_line, file_name):
        """Classify the given header name and return the corresponding index.

        Keyword arguments:
        config_file -- Config object
        include_line -- regex Match object for include line
        file_name -- file name string

        Returns header classification index.
        """
        # Process override regexes
        for idx in range(5):
            if self.override_regexes[idx].search(include_line.group("name")):
                return idx

        include_name = os.path.basename(include_line.group("name"))
        include_base, include_ext = os.path.splitext(include_name)
        file_base, file_ext = os.path.splitext(file_name)

        # Is related if include has same base name as file name and file has a
        # source extension
        include_is_related = include_base == file_base and (
            config_file.is_c_src_file(file_name)
            or config_file.is_cpp_src_file(file_name)
        )

        if include_is_related:
            return 0
        elif include_line.group("name") in self.c_std:
            return 1
        elif self.c_sys_regex.search(include_line.group("header")):
            return 1
        elif self.cpp_std_regex.search(include_line.group("header")):
            return 2
        elif include_line.group("open_bracket") == "<":
            return 3
        elif self.include_is_header(config_file, file_name, include_name):
            return 4
        else:
            return -1

    @staticmethod
    def include_is_header(config_file, file_name, include_name):
        """Return True if include name has header extension.

        Keyword arguments:
        config_file -- Config object
        file_name -- file name string
        include_name -- include name string

        Returns true if include name is a header.
        """
        if not config_file.is_header_file(include_name):
            print(
                "Error: "
                + file_name
                + ": include '"
                + include_name
                + "' has extension not in header list"
            )
            return False
        else:
            return True

    @staticmethod
    def rebuild_include(name_match, group_number):
        """Adds appropriate brackets around include name and "#include" before
        that based on group number.

        Keyword arguments:
        name_match -- include name's regex Match object
        group_number -- include classification index

        Returns include name with approriate brackets and "#include" prefix.
        """
        if 1 <= group_number <= 3:
            output = (
                "#include <"
                + name_match.group("name")
                + ">"
                + name_match.group("postfix")
            )
        else:
            output = (
                '#include "'
                + name_match.group("name")
                + '"'
                + name_match.group("postfix")
            )

        if name_match.group("comment"):
            return name_match.group("comment") + output
        else:
            return output

    @staticmethod
    def dedup_list(seq):
        """Remove duplicates from list while preserving order.

        Keyword arguments:
        seq -- list from which to deduplicate entries

        Returns deduplicated list.
        """
        seen = set()
        seen_add = seen.add
        return [x for x in seq if not (x in seen or seen_add(x))]

    def write_headers(self, includes, ifdef_blocks=None):
        """Write out includes from sets.

        Keyword arguments:
        includes -- list of include strings
        ifdef_blocks -- list of list of #ifdef blocks (default empty 2D list)

        Returns list of output lines.
        """
        if ifdef_blocks is None:
            ifdef_blocks = [[], [], [], [], []]
        output_list = []

        for i in range(5):
            if includes[i]:
                sublist = sorted(includes[i], key=lambda include: include.group("name"))
                str_list = [self.rebuild_include(x, i) for x in sublist]
                output_list.extend(self.dedup_list(str_list))
                output_list.append("")  # Delimits groups of includes

            # #ifdef blocks go after other includes
            if ifdef_blocks[i]:
                # Remove newline from last #endif
                output_list.append(self.linesep.join(ifdef_blocks[i]).rstrip())
                output_list.append("")  # Delimits groups of #ifdef blocks
        if output_list:
            del output_list[-1]  # Remove last newline
        return output_list

    def header_sort(self, config_file, lines_list, file_name, start, end, ifdef_level):
        """Recursively parses within #ifdef blocks for header includes and sorts
        them.

        Keyword arguments:
        config_file -- Config object
        lines_list -- list of file contents delimited by line separator
        file_name -- file name string
        start -- starting line for recursion step
        end -- ending line for recursion step
        ifdef_level -- current recursion depth of #ifdefs

        Returns tuple of the following:
          sorted output
          list of bools indicating header category presence within #ifdef block
          the index of the last line processed
          whether all header includes had header extension
        """
        output_list = []

        # Using sets here eliminates duplicate includes
        includes = [[], [], [], [], []]

        # List of bools indicating header category presence within #ifdef block
        includes_present = [False, False, False, False, False]

        ifdef_blocks = [[], [], [], [], []]

        i = start
        while i < end:
            if "#ifdef" in lines_list[i]:
                ifdef_count = 1
                for j in range(i + 1, end):
                    if "#ifdef" in lines_list[j]:
                        ifdef_count += 1
                    elif "#endif" in lines_list[j]:
                        ifdef_count -= 1
                    if ifdef_count == 0:
                        # Add #ifdef or #else line
                        ifdef = lines_list[i] + self.linesep

                        suboutput, inc_present, idx, valid_headers = self.header_sort(
                            config_file,
                            lines_list,
                            file_name,
                            i + 1,
                            j,
                            ifdef_level + 1,
                        )
                        i = j

                        # If header failed to classify, return failure
                        if not valid_headers:
                            return output_list, inc_present, i, False

                        if suboutput:
                            ifdef += self.linesep.join(suboutput) + self.linesep

                        # Add #endif line
                        ifdef += lines_list[j]

                        # #endif gets an extra line separator
                        if "#endif" in lines_list[j]:
                            ifdef += self.linesep

                        for k in range(len(includes_present)):
                            includes_present[k] |= inc_present[k]

                        # If there is only one type of include present
                        if sum(inc_present) == 1:
                            ifdef_idx = 0
                            for k in range(len(inc_present)):
                                if inc_present[k]:
                                    ifdef_idx = k
                                    break

                            # Only one type of include, so add it to sorts
                            ifdef_blocks[ifdef_idx].append(ifdef)
                        else:
                            # Treat #ifdef as barrier and flush includes
                            output_list.extend(self.write_headers(includes))
                            if output_list and not output_list[-1].endswith(
                                self.linesep
                            ):
                                # Only add newline if content exists in
                                # output_list that doesn't end in a newline,
                                # which needs to be delimited from the #ifdef
                                output_list.append("")
                            includes = [[], [], [], [], []]

                            output_list.append(ifdef)
                        break
            elif "#include" in lines_list[i]:
                include_line = self.header_regex.search(lines_list[i])
                if not include_line:
                    # Write headers and #ifdef blocks if found
                    written = self.write_headers(includes, ifdef_blocks)
                    if written:
                        output_list.extend(written)
                    return output_list, includes_present, i, True

                if "NOLINT" in lines_list[i]:
                    # NOLINT is a barrier, so flush includes
                    written = self.write_headers(includes)
                    if written:
                        output_list.append(self.linesep.join(written))
                        includes = [[], [], [], [], []]

                    # NOLINT line will have newlines above and below, unless it
                    # is the first line being processed.
                    if output_list and output_list[-1] != "":
                        output_list.append("")
                    output_list.append(lines_list[i])
                    output_list.append("")

                    i += 1
                    continue

                # Insert header into appropriate list
                idx = self.classify_header(config_file, include_line, file_name)
                if idx != -1:
                    if include_line.group("comment"):
                        print(
                            "Warning: "
                            + file_name
                            + ": include '"
                            + include_line.group("name")
                            + "' is commented out"
                        )
                    includes[idx].append(include_line)
                    includes_present[idx] = True
                else:
                    # If header failed to classify, return failure
                    return output_list, includes_present, i, False
            elif ifdef_level > 0 and lines_list[i] != "":
                # Non-preprocessor statements within a #ifdef block don't mark
                # the end of header include processing, but act as barrier for
                # sorting.

                # Dump currently collected header includes to file.
                # includes_present isn't reset as well here because that info is
                # still valid and necessary for sorting.
                written = self.write_headers(includes)
                if written:
                    output_list.append(self.linesep.join(written))
                    includes = [[], [], [], [], []]

                output_list.append(lines_list[i])
            elif lines_list[i].strip() != "":
                # Write headers and #ifdef blocks if found
                written = self.write_headers(includes, ifdef_blocks)
                if written:
                    output_list.extend(written)

                return output_list, includes_present, i, True
            i += 1

        # Write headers and #ifdef blocks if found
        written = self.write_headers(includes, ifdef_blocks)
        if written:
            output_list.extend(written)

        return output_list, includes_present, i, True

    def run_pipeline(self, config_file, name, lines):
        self.override_regexes = []

        # Compile include sorting override regexes
        for group in [
            "includeRelated",
            "includeCSys",
            "includeCppSys",
            "includeOtherLibs",
            "includeProject",
        ]:
            regex_str = config_file.regex(group)
            self.override_regexes.append(regex.compile(regex_str))

        self.linesep = super().get_linesep(lines)

        file_name = os.path.basename(name)

        lines_list = lines.splitlines()

        # Write lines from beginning of file to headers
        i = 0
        while i < len(lines_list) and (
            "#ifdef" not in lines_list[i] and "#include" not in lines_list[i]
        ):
            i += 1
        output_list = lines_list[0:i]

        suboutput, inc_present, idx, valid_headers = self.header_sort(
            config_file, lines_list, file_name, i, len(lines_list), 0
        )
        i = idx

        # If header failed to classify, return failure
        if not valid_headers:
            return lines, False

        if suboutput:
            output_list.extend(suboutput)

        # Remove extra empty lines from end of includes
        while len(output_list) > 0 and output_list[-1].rstrip() == "":
            del output_list[-1]  # Remove last newline

        # Remove possible extra newline from #endif
        if len(output_list) > 0:
            output_list[-1] = output_list[-1].rstrip()
            output_list.append("")

        # Write rest of file
        output_list.extend(lines_list[i:])

        output = self.linesep.join(output_list).rstrip() + self.linesep
        return output, True
